"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createNoUnsafeAssignmentRule = void 0;
/* eslint-disable functional/prefer-immutable-types */
const utils_1 = require("@typescript-eslint/utils");
const typescript_1 = require("typescript");
const tsutils_1 = require("tsutils");
const common_1 = require("./common");
const maxDepth = 100;
const isSignatureAssignable = (destinationNode, sourceNode, destinationSignature, sourceSignature, checker) => {
    const returnTypeIsAssignable = checker.isTypeAssignableTo(sourceSignature.getReturnType(), destinationSignature.getReturnType());
    const allParametersAreAssignable = destinationSignature.parameters.every((destinationParameter, index) => {
        const sourceParameter = sourceSignature.parameters[index];
        // eslint-disable-next-line functional/no-conditional-statements
        if (sourceParameter === undefined) {
            return false;
        }
        // source and destination parameter reversed here due to contravariance.
        return checker.isTypeAssignableTo(checker.getTypeOfSymbolAtLocation(destinationParameter, destinationNode), checker.getTypeOfSymbolAtLocation(sourceParameter, sourceNode));
    });
    return returnTypeIsAssignable && allParametersAreAssignable;
};
const assignableSignaturePairs = (destinationNode, sourceNode, destinationSignatures, sourceSignatures, checker) => {
    return sourceSignatures.flatMap((sourceSignature) => destinationSignatures
        .filter((destinationSignature) => isSignatureAssignable(destinationNode, sourceNode, sourceSignature, destinationSignature, checker))
        .map((destinationSignature) => ({
        sourceSignature,
        destinationSignature,
    })));
};
const createNoUnsafeAssignmentRule = (unsafePropertyAssignmentFunc, unsafeIndexAssignmentFunc) => (context
// eslint-disable-next-line sonarjs/cognitive-complexity
) => {
    const parserServices = utils_1.ESLintUtils.getParserServices(context);
    const checker = parserServices.program.getTypeChecker();
    const isUnsafeIndexAssignment = (indexKind, destinationNode, sourceNode, destinationType, sourceType, checker, depth, seenTypes) => {
        const isUnsafe = unsafeIndexAssignmentFunc(indexKind, destinationType, sourceType, checker);
        const destinationIndexType = indexKind === typescript_1.IndexKind.Number
            ? destinationType.getNumberIndexType()
            : destinationType.getStringIndexType();
        const sourceIndexType = indexKind === typescript_1.IndexKind.Number
            ? sourceType.getNumberIndexType()
            : sourceType.getStringIndexType();
        // This is unsafe if...
        return (
        // this particular rule considers the index assignment unsafe, or
        isUnsafe ||
            // the index types are considered unsafe themselves (recursively).
            (destinationIndexType !== undefined &&
                sourceIndexType !== undefined &&
                isUnsafeAssignment(destinationNode, sourceNode, destinationIndexType, sourceIndexType, checker, depth, seenTypes)));
    };
    const isUnsafePropertyAssignmentRec = (destinationNode, sourceNode, 
    // eslint-disable-next-line @typescript-eslint/ban-types
    destinationProperty, 
    // eslint-disable-next-line @typescript-eslint/ban-types
    sourceProperty, checker, depth, seenTypes) => {
        const destinationPropertyType = checker.getTypeOfSymbolAtLocation(destinationProperty, destinationNode);
        const sourcePropertyType = checker.getTypeOfSymbolAtLocation(sourceProperty, sourceNode);
        return isUnsafeAssignment(destinationNode, sourceNode, destinationPropertyType, sourcePropertyType, checker, depth, seenTypes);
    };
    const isUnsafePropertyAssignment = (destinationNode, sourceNode, destinationType, sourceType, checker, depth, seenTypes) => {
        // eslint-disable-next-line functional/no-conditional-statements
        if ((checker.isArrayType(destinationType) ||
            checker.isArrayType(sourceType)) &&
            !(0, tsutils_1.isUnionOrIntersectionType)(destinationType) &&
            !(0, tsutils_1.isUnionOrIntersectionType)(sourceType)) {
            // For performance reasons we avoid checking every array property for unsafe assignment.
            //
            // If one of these arrays is readonly and the other mutable, it will be caught by `isUnsafeIndexAssignment`,
            // so we needn't compare each individual property.
            //
            // TODO what about tuples?
            // TODO should we ignore other built-in types like strings, symbols and numbers for performance too?
            return false;
        }
        return destinationType.getProperties().some((destinationProperty) => {
            const sourceProperty = sourceType.getProperty(destinationProperty.name);
            const isUnsafe = unsafePropertyAssignmentFunc(destinationProperty, sourceProperty, destinationType, sourceType, checker);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafe) {
                return true;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if (sourceProperty === undefined) {
                return false;
            }
            const nextSeenTypes = seenTypes.concat({
                destinationType,
                sourceType,
                destinationRecursionIdentity: checker.getRecursionIdentity(destinationType),
                sourceRecursionIdentity: checker.getRecursionIdentity(sourceType),
            });
            return isUnsafePropertyAssignmentRec(destinationNode, sourceNode, destinationProperty, sourceProperty, checker, depth, nextSeenTypes);
        });
    };
    const isUnsafeAssignment = (destinationNode, sourceNode, rawDestinationType, rawSourceType, checker, depth, seenTypes = []) => {
        // eslint-disable-next-line functional/no-conditional-statements
        if (rawDestinationType === rawSourceType) {
            // Never unsafe if the types are equal.
            return false;
        }
        // eslint-disable-next-line functional/no-conditional-statements
        if (depth >= maxDepth) {
            return false;
        }
        const nextSeenTypes = seenTypes.concat({
            destinationType: rawDestinationType,
            sourceType: rawSourceType,
            destinationRecursionIdentity: checker.getRecursionIdentity(rawDestinationType),
            sourceRecursionIdentity: checker.getRecursionIdentity(rawSourceType),
        });
        const typePairs = (0, common_1.assignableTypePairs)(rawDestinationType, rawSourceType, checker);
        const objectTypePairs = typePairs.filter((tp) => (0, tsutils_1.intersectionTypeParts)(tp.destinationType).some(tsutils_1.isObjectType));
        const isUnsafeFunctionAssignment = (typePairs) => {
            return typePairs.some(({ sourceType, destinationType }) => {
                const nextSeenTypesWithPair = nextSeenTypes.concat({
                    destinationType,
                    sourceType,
                    destinationRecursionIdentity: checker.getRecursionIdentity(destinationType),
                    sourceRecursionIdentity: checker.getRecursionIdentity(sourceType),
                });
                const sourceCallSignatures = (0, tsutils_1.getCallSignaturesOfType)(sourceType);
                const destinationCallSignatures = (0, tsutils_1.getCallSignaturesOfType)(destinationType);
                const signaturePairs = assignableSignaturePairs(destinationNode, sourceNode, destinationCallSignatures, sourceCallSignatures, checker);
                return signaturePairs.some(({ destinationSignature, sourceSignature }) => {
                    // eslint-disable-next-line functional/functional-parameters
                    const isUnsafeParameterAssignment = () => {
                        return destinationSignature
                            .getParameters()
                            .some((destinationParameterSymbol, index) => {
                            const destinationParameterType = checker.getTypeOfSymbolAtLocation(destinationParameterSymbol, destinationNode);
                            const sourceParameterSymbol = sourceSignature.getParameters()[index];
                            // eslint-disable-next-line functional/no-conditional-statements
                            if (sourceParameterSymbol === undefined) {
                                return false;
                            }
                            const sourceParameterType = checker.getTypeOfSymbolAtLocation(sourceParameterSymbol, sourceNode);
                            // Source and destination swapped here - function parameters are contravariant.
                            return isUnsafeAssignment(sourceNode, destinationNode, sourceParameterType, destinationParameterType, checker, depth + 1, nextSeenTypesWithPair);
                        });
                    };
                    const sourceReturnType = sourceSignature.getReturnType();
                    const sourceRecursionIdentity = checker.getRecursionIdentity(sourceReturnType);
                    const destinationReturnType = destinationSignature.getReturnType();
                    const destinationRecursionIdentity = checker.getRecursionIdentity(destinationType);
                    // This is an unsafe assignment if...
                    return (
                    // we're not in an infinitely recursive type,
                    seenTypes.every((t) => t.destinationRecursionIdentity !==
                        destinationRecursionIdentity &&
                        t.sourceRecursionIdentity !== sourceRecursionIdentity) &&
                        // and the types we're assigning from and to are different,
                        destinationType !== sourceType &&
                        // and the return types of the functions are unsafe assignment,
                        (isUnsafeAssignment(destinationNode, sourceNode, destinationReturnType, sourceReturnType, checker, depth + 1, nextSeenTypesWithPair) ||
                            // or the parameter types of the functions are unsafe assignment.
                            isUnsafeParameterAssignment()));
                });
            });
        };
        const inUnsafeObjectAssignment = (objectTypePairs) => objectTypePairs.some(({ sourceType, destinationType }) => {
            const sourceRecursionIdentity = checker.getRecursionIdentity(sourceType);
            const destinationRecursionIdentity = checker.getRecursionIdentity(destinationType);
            const nextSeenTypesWithPair = nextSeenTypes.concat({
                destinationType,
                sourceType,
                destinationRecursionIdentity: destinationRecursionIdentity,
                sourceRecursionIdentity: sourceRecursionIdentity,
            });
            // This is an unsafe assignment if...
            return (
            // we're not in an infinitely recursive type,
            seenTypes.every((t) => t.destinationRecursionIdentity !==
                destinationRecursionIdentity &&
                t.sourceRecursionIdentity !== sourceRecursionIdentity) &&
                // and the types we're assigning from and to are different,
                // TODO this seems to be required to prevent a hang in https://github.com/oaf-project/oaf-react-router
                // Need to work out why and formulate a test to reproduce
                destinationType !== sourceType &&
                // and we're either:
                // unsafe string index assignment, or
                (isUnsafeIndexAssignment(typescript_1.IndexKind.String, destinationNode, sourceNode, destinationType, sourceType, checker, depth + 1, nextSeenTypesWithPair) ||
                    // unsafe number index assignment, or
                    isUnsafeIndexAssignment(typescript_1.IndexKind.Number, destinationNode, sourceNode, destinationType, sourceType, checker, depth + 1, nextSeenTypesWithPair) ||
                    // unsafe property assignment.
                    isUnsafePropertyAssignment(destinationNode, sourceNode, destinationType, sourceType, checker, depth + 1, nextSeenTypesWithPair)));
        });
        return (inUnsafeObjectAssignment(objectTypePairs) ||
            isUnsafeFunctionAssignment(typePairs));
    };
    // Special handling for array methods that return mutable arrays but that
    // we know are shallow copies and therefore safe to have their result
    // assigned to a readonly array.
    const isSafeAssignmentFromArrayMethod = (sourceExpression, destinationNode, sourceNode, destinationType, sourceType, checker, depth, parserServices) => {
        const safeArrayMethods = [
            "filter",
            "map",
            "concat",
            "flatMap",
            "flat",
            "slice",
        ];
        // Arrays have number index types. This gives us access to the type within the array.
        const destinationIndexType = destinationType.getNumberIndexType();
        const sourceIndexType = sourceType.getNumberIndexType();
        // eslint-disable-next-line functional/no-conditional-statements, sonarjs/prefer-single-boolean-return
        if (
        // We are assigning to and from arrays
        checker.isArrayType(destinationType) &&
            checker.isArrayType(sourceType) &&
            destinationIndexType !== undefined &&
            sourceIndexType !== undefined &&
            // and the assignment is from calling a member (obj.method(...))
            sourceExpression.type === utils_1.AST_NODE_TYPES.CallExpression &&
            sourceExpression.callee.type === utils_1.AST_NODE_TYPES.MemberExpression &&
            // and the thing being called is an array
            // (so we can avoid permitting calls to array methods on other types)
            checker.isArrayType(checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(sourceExpression.callee.object))) &&
            // and the method being called is an identifier that we can match on like "concat"
            sourceExpression.callee.property.type === utils_1.AST_NODE_TYPES.Identifier &&
            // and it's not computed (obj["method"])
            // TODO: support computed properties like myArray["concat"]?
            !sourceExpression.callee.computed &&
            // and the method being called is one that we know is safe to assign to a readonly array
            safeArrayMethods.includes(sourceExpression.callee.property.name) &&
            // and the types within the source and destination array are themselves safe to assign
            // (avoid this issue: https://github.com/danielnixon/eslint-plugin-total-functions/issues/730)
            !isUnsafeAssignment(destinationNode, sourceNode, destinationIndexType, sourceIndexType, checker, depth)) {
            return true;
        }
        return false;
    };
    return {
        // eslint-disable-next-line functional/no-return-void
        TSTypeAssertion: (node) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.typeAnnotation.type === utils_1.AST_NODE_TYPES.TSTypeReference &&
                node.typeAnnotation.typeName.type === utils_1.AST_NODE_TYPES.Identifier &&
                node.typeAnnotation.typeName.name === "const") {
                // Always allow `as const`.
                return;
            }
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = destinationNode.expression;
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringTSTypeAssertion",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        TSAsExpression: (node) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.typeAnnotation.type === utils_1.AST_NODE_TYPES.TSTypeReference &&
                node.typeAnnotation.typeName.type === utils_1.AST_NODE_TYPES.Identifier &&
                node.typeAnnotation.typeName.name === "const") {
                // Always allow `as const`.
                return;
            }
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = destinationNode.expression;
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringTSAsExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        VariableDeclaration: (node) => {
            // eslint-disable-next-line functional/no-expression-statements, functional/no-return-void
            node.declarations.forEach((declaration) => {
                // eslint-disable-next-line functional/no-conditional-statements
                if (declaration.id.type === utils_1.AST_NODE_TYPES.Identifier &&
                    declaration.id.typeAnnotation === undefined) {
                    // If there is no type annotation then there's no risk of unsafe assignment.
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statements
                if (declaration.init === null) {
                    // If there is no type annotation then there's no risk of unsafe assignment.
                    return;
                }
                const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(declaration.id);
                const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(declaration.init);
                const leftType = checker.getTypeAtLocation(leftTsNode);
                const rightType = checker.getTypeAtLocation(rightTsNode);
                // eslint-disable-next-line functional/no-conditional-statements
                if (isSafeAssignmentFromArrayMethod(declaration.init, leftTsNode, rightTsNode, leftType, rightType, checker, 0, parserServices)) {
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statements
                if (isUnsafeAssignment(leftTsNode, rightTsNode, leftType, rightType, checker, 0)) {
                    // eslint-disable-next-line functional/no-expression-statements
                    context.report({
                        node: node,
                        messageId: "errorStringVariableDeclaration",
                    });
                }
            });
        },
        // eslint-disable-next-line functional/no-return-void
        AssignmentExpression: (node) => {
            const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.left);
            const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.right);
            const leftType = checker.getTypeAtLocation(leftTsNode);
            const rightType = checker.getTypeAtLocation(rightTsNode);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isSafeAssignmentFromArrayMethod(node.right, leftTsNode, rightTsNode, leftType, rightType, checker, 0, parserServices)) {
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(leftTsNode, rightTsNode, leftType, rightType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringAssignmentExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        ReturnStatement: (node) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.argument === null) {
                return;
            }
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            // eslint-disable-next-line functional/no-conditional-statements
            if (tsNode.expression === undefined) {
                return;
            }
            const destinationType = checker.getContextualType(tsNode.expression);
            // eslint-disable-next-line functional/no-conditional-statements
            if (destinationType === undefined) {
                return;
            }
            const sourceType = checker.getTypeAtLocation(tsNode.expression);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isSafeAssignmentFromArrayMethod(node.argument, tsNode.expression, tsNode.expression, destinationType, sourceType, checker, 0, parserServices)) {
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(tsNode.expression, tsNode.expression, destinationType, sourceType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringArrowFunctionExpression",
                });
            }
        },
        // TODO fix this copypasta between YieldExpression and ReturnStatement
        // eslint-disable-next-line functional/no-return-void
        YieldExpression: (node) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.argument === undefined) {
                return;
            }
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            // eslint-disable-next-line functional/no-conditional-statements
            if (tsNode.expression === undefined) {
                return;
            }
            const destinationType = checker.getContextualType(tsNode.expression);
            // eslint-disable-next-line functional/no-conditional-statements
            if (destinationType === undefined) {
                return;
            }
            const sourceType = checker.getTypeAtLocation(tsNode.expression);
            // eslint-disable-next-line functional/no-conditional-statements
            if (isSafeAssignmentFromArrayMethod(node.argument, tsNode.expression, tsNode.expression, destinationType, sourceType, checker, 0, parserServices)) {
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(tsNode.expression, tsNode.expression, destinationType, sourceType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringArrowFunctionExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        ArrowFunctionExpression: (node) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.returnType === undefined) {
                return;
            }
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node.returnType.typeAnnotation);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = parserServices.esTreeNodeToTSNodeMap.get(node.body);
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statements
            if (node.body.type !== utils_1.AST_NODE_TYPES.BlockStatement &&
                isSafeAssignmentFromArrayMethod(node.body, destinationNode, sourceNode, destinationType, sourceType, checker, 0, parserServices)) {
                return;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker, 0)) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node.body,
                    messageId: "errorStringArrowFunctionExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        CallExpression: (node) => {
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            // eslint-disable-next-line functional/no-expression-statements, functional/no-return-void
            tsNode.arguments.forEach((argument, i) => {
                const argumentType = checker.getTypeAtLocation(argument);
                const paramType = checker.getContextualType(argument);
                // eslint-disable-next-line functional/no-conditional-statements
                if (paramType !== undefined &&
                    isUnsafeAssignment(argument, argument, paramType, argumentType, checker, 0)) {
                    // eslint-disable-next-line functional/no-expression-statements
                    context.report({
                        node: node.arguments[i] ?? node,
                        messageId: "errorStringCallExpression",
                    });
                }
            });
        },
    };
};
exports.createNoUnsafeAssignmentRule = createNoUnsafeAssignmentRule;
